from dataclasses import dataclass
@dataclass
class Group:
    """This is a basic group function to store the blocks, the get and set items have been added for backwards compatibility"""
    a:int
    b:int
    c:int
    d:int
    is_version_b: bool=False #should be none if theres no group set
    def to_list(self):
        return [self.a, self.b, self.c, self.d]
    def __iter__(self):
        return self.to_list()
    def __getitem__(self, key):
        return getattr(self, ["a","b","c","d"][key], None)
    def __setitem__(self, key, value):
        setattr(self, ["a","b","c","d"][key], value)


class GroupInterface:
    def getPS(text: str, full=None): #variable left for backwards compatibility
        if not full is None: print("Please update your code for the librds 1.2+ version. (getPS, this will be removed on version 2)")
        if len(text) > 8: text = text[8:]
        return text.ljust(8), 4
    def getRT(text: str,full:bool=False):
        if len(text) > 64: text = text[64:]
        else: text += "\r" # http://www.interactive-radio-system.com/docs/EN50067_RDS_Standard.pdf page 26
        if not full:
            while len(text) % 4: # if we don't have text to equally spread across 4 charcter parts then we add padding
                text = text + " "
            segments = 0
            for _ in range(len(text)):
                segments = segments + 0.25 # 0.25*4 = 1
            return text, int(segments)
        else:
            return text.ljust(64), 16
    def getPTYN(text: str,full=None): #variable left for backwards compatibility
        if not full is None: print("Please update your code for the librds 1.2+ version. (getPTYN, this will be removed on version 2)")
        if len(text) > 8: text = text[8:]
        return text.ljust(8), 2


from enum import Enum, IntEnum

def get_from_list(input:list,index:int,default=None):
    """This is a simple function to remove index errors from the group generators"""
    try:
        return input[index]
    except IndexError:
        return default

class Groups(Enum):
    PS = 0
    PS_B = 1
    RT = 2
    RT_B = 3
    PTYN = 4
    ECC = 5
    LIC = 6
    TDA = 7
    TDA_B = 8
    IN_HOUSE = 9
    IN_HOUSE_B = 10

class GroupSequencer:
    """you can use this code to sequence though anything"""
    def __init__(self, sequence:list[IntEnum]) -> None:
        self.cur_idx = 1
        self.sequence = sequence
    def get_next(self):
        if len(self.sequence) == 0: return
        if self.cur_idx > len(self.sequence)-1: self.cur_idx = 1
        prev = self.sequence[self.cur_idx-1]
        self.cur_idx += 1
        return prev
    def change_sequence(self, sequence:list[IntEnum]):
       self.sequence = sequence
       self.cur_idx = 1
    def __len__(self):
        return len(self.sequence)



from .interface import GroupInterface
from .comfort import Groups, GroupSequencer
from .af import AF_Bands, AlternativeFrequencyEntry
from .generator import GroupGenerator, Group
from .decoder import GroupDecoder
librds_version = 1.36
pyrds_version = 1.2



from .group import Group
from .comfort import get_from_list
from .af import AlternativeFrequency
from .charset import *

class GroupGenerator:
    def basic(pi:int, tp: bool=False, pty: int=0):
        """This function will generate a basic block structure which includes the PI, TP and PTY, this shouldn't be sent by itself to a decoder"""
        return Group(
            pi & 0xFFFF, #A
            (int(tp) << 10 | pty << 5) & 0xFFFF, #B
            0,0, #C
            None #is_b
        )
    def ps(blocks:Group,ps_text:str,segment:int,ms:bool=True,ta:bool=False,di:int=1,block2:int=0):
        """This function will generate a 0A group which includes TA MS DI AF and text data"""
        if segment > 3: raise Exception("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            ( blocks.b | int(ta) << 4 | int(ms) << 3 | ( ( di >> (3-segment) ) << 2) | segment ) & 0xFFFF,
            int(block2) & 0xFFFF or AlternativeFrequency.get_no_af(),
            (RDSCharset.translate(get_from_list(ps_text,segment*2," "))<<8 | RDSCharset.translate(get_from_list(ps_text,segment*2+1," "))) & 0xFFFF, #low byte + high byte as we're using utf-8, and we conver 2 8 bit numbers to a single 16 bit one
            False
        )
    def ps_b(blocks:Group,ps_text:str,segment:int,ms:bool=True,ta:bool=False,di:int=1):
        """This function will generate a 0B group which is very similiar to a 0A group but no AF"""
        if segment > 3: raise Exception("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            ( blocks.b | int(ta) << 4 | int(ms) << 3 | ( ( di >> (3-segment) ) << 2) | segment ) & 0xFFFF,
            blocks.a & 0xFFFF,
            (RDSCharset.translate(get_from_list(ps_text,segment*2," "))<<8 | RDSCharset.translate(get_from_list(ps_text,segment*2+1," "))) & 0xFFFF,
            True
        )
    def ecc(blocks:Group, ecc: int):
        """This function will generate a 1A group"""
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | 1 << 12) & 0xFFFF,
            (blocks.c | ecc) & 0xFFFF,
            blocks.d & 0xFFFF,
            False
        )
    def lic(blocks:Group, lic: int):
        """This function will generate a 1A group"""
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | 1 << 12) & 0xFFFF,
            (blocks.c | (lic | 0x3000)) & 0xFFFF,
            blocks.d & 0xFFFF,
            False
        )
    def rt(blocks:Group,rt_text:str,segment:int,ab:bool=False):
        """This function will generate a 2A group"""
        if segment > 15: raise Exception("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | 2 << 12 | int(ab) << 4 | segment) & 0xFFFF,
            (RDSCharset.translate(get_from_list(rt_text,segment*4+0," "))<<8 | RDSCharset.translate(get_from_list(rt_text,segment*4+1," "))) & 0xFFFF,
            (RDSCharset.translate(get_from_list(rt_text,segment*4+2," "))<<8 | RDSCharset.translate(get_from_list(rt_text,segment*4+3," "))) & 0xFFFF,
            False
        )
    def rt_b(blocks:Group,rt_text:str,segment:int,ab:bool=False):
        """This function will generate a 2B group"""
        if segment > 15: raise Exception("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            (0x0800 | blocks.a | 2 << 12 | int(ab) << 4 | segment) & 0xFFFF,
            blocks.a & 0xFFFF,
            (RDSCharset.translate(get_from_list(rt_text,segment*4+0," "))<<8 | RDSCharset.translate(get_from_list(rt_text,segment*4+1," "))) & 0xFFFF,
            True
        )
    def tda(blocks:Group, channel:int,data:list[int],segment:int):
        """This function will generate a 5A group"""
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | 5 << 12 | (channel & 0x001F)) & 0xFFFF,
            (get_from_list(data,segment+0,0)<<8 | get_from_list(data,segment+1,0)) & 0xFFFF,
            (get_from_list(data,segment+2,0)<<8 | get_from_list(data,segment+3,0)) & 0xFFFF,
            False
        )
    def tda_b(blocks:Group, channel:int,data:list[int],segment:int):
        """This function will generate a 5B group"""
        return Group(
            blocks.a & 0xFFFF,
            (0x0800 | blocks.b | 5 << 12 | (channel & 0x001F)) & 0xFFFF,
            blocks.a & 0xFFFF,
            (get_from_list(data,segment+0,0)<<8 | get_from_list(data,segment+1,0)) & 0xFFFF,
            True
        )
    def in_house(blocks:Group, data:list[int]):
        """This function will generate a 6A group"""
        return Group(
            blocks.a & 0xFFFF,
            (((blocks.b | 6 << 12) & ~0b11111) | (get_from_list(data,0,0) & 0b11111)) & 0xFFFF,
            get_from_list(data,1,0) & 0xFFFF,
            get_from_list(data,2,0) & 0xFFFF,
            False
        )
    def in_house_b(blocks:Group, data:list[int]):
        """This function will generate a 6B group"""
        return Group(
            blocks.a & 0xFFFF,
            (((0x0800 | blocks.b | 6 << 12) & ~0b11111) | (get_from_list(data,0,0) & 0b11111)) & 0xFFFF,
            blocks.a & 0xFFFF,
            get_from_list(data,1,0) & 0xFFFF,
            True
        )
    def ptyn(blocks:Group, ptyn_text:str, segment:int,ab:bool=False):
        """This function will generate a 10A group"""
        if segment > 1: raise Exception("Segment limit")
        return Group(
            blocks.a & 0xFFFF,
            (blocks.b | 10 << 12 | (int(ab) << 4) | segment) & 0xFFFF,
            (RDSCharset.translate(get_from_list(ptyn_text,segment*4+0," "))<<8 | RDSCharset.translate(get_from_list(ptyn_text,segment*4+1," "))) & 0xFFFF,
            (RDSCharset.translate(get_from_list(ptyn_text,segment*4+2," "))<<8 | RDSCharset.translate(get_from_list(ptyn_text,segment*4+3," "))) & 0xFFFF,
            False
        )



from enum import IntEnum
from .charset import RDSCharsetDecode
from .group import Group, dataclass

@dataclass
class Details:
    details=True
@dataclass
class PSDetails(Details):
    segment: int
    di: int
    ms: bool
    ta: bool
    text: str
@dataclass
class RTDetails(Details):
    segment: int
    ab: int
    text: str
@dataclass
class PTYNDetails(Details):
    segment: int
    ab: int
    text: str
@dataclass
class ECCLICDetails(Details):
    data:int
    is_lic:bool
@dataclass
class DecodedGroup:
    raw_group:Group
    pi: int
    tp: bool
    pty: int
    group: IntEnum
    details:Details
class GroupDecoder:
    def _decode_0(self, group: Group, dgroup: DecodedGroup):
        segment = group.b & 3
        ta = (group.b >> 4) & 1
        ms = (group.b >> 3) & 1
        if segment == 3:
            di = (group.b >> 2) & 9
        else:
            di = 0
        details = PSDetails(segment,di,bool(ms),bool(ta),"")

        char_1 = (group.d >> 8) & 0xFF
        char_2 = group.d & 0xFF
        details.text += RDSCharsetDecode.translate(char_1)
        details.text += RDSCharsetDecode.translate(char_2)

        dgroup.details = details
        return group, dgroup
    def _decode_2(self, group: Group, dgroup: DecodedGroup):
        dgroup.group = 2
        segment = group.b & 15
        ab = (group.b >> 4) & 1
        details = RTDetails(segment,ab,"")

        if not group.is_version_b:
            char_1 = (group.c >> 8) & 0xFF
            char_2 = group.c & 0xFF
            char_3 = (group.d >> 8) & 0xFF
            char_4 = group.d & 0xFF
            details.text += RDSCharsetDecode.translate(char_1)
            details.text += RDSCharsetDecode.translate(char_2)
            details.text += RDSCharsetDecode.translate(char_3)
            details.text += RDSCharsetDecode.translate(char_4)
        else:
            char_1 = (group.d >> 8) & 0xFF
            char_2 = group.d & 0xFF
            details.text += RDSCharsetDecode.translate(char_1)
            details.text += RDSCharsetDecode.translate(char_2)

        dgroup.details = details
        return group, dgroup
    def _decode_1(self, group:Group, dgroup:DecodedGroup):
        dgroup.group = 1
        tmp = (group.c & 0xFF)
        details = ECCLICDetails(0,False)
        if tmp < 0x46:
            details.is_lic = True
        details.data = tmp
        dgroup.details = details
        return group, dgroup
    def _decode_10(self, group:Group, dgroup:DecodedGroup):
        dgroup.group = 10
        segment = group.b & 3
        ab = (group.b >> 4) & 1
        details = PTYNDetails(segment,ab,"")
        char_1 = (group.c >> 8) & 0xFF
        char_2 = group.c & 0xFF
        char_3 = (group.d >> 8) & 0xFF
        char_4 = group.d & 0xFF
        details.text += RDSCharsetDecode.translate(char_1)
        details.text += RDSCharsetDecode.translate(char_2)
        details.text += RDSCharsetDecode.translate(char_3)
        details.text += RDSCharsetDecode.translate(char_4)
        dgroup.details = details
        return group, dgroup
    def decode(self, group: Group):
        out = DecodedGroup(group,group.a,False,0,0,None)
        out.pty = ((group.b >> 5) & 31)
        out.tp = bool((group.b >> 10) & 1)
        gr = (group.b >> 12) & ((1 << 12) - 1)
        if group.is_version_b and gr != 0: gr -= 1
        match gr:
            case 0:
                group, out = self._decode_0(group, out)
            case 2:
                group, out = self._decode_2(group, out)
            case 1:
                group, out = self._decode_1(group, out)
            case 10:
                group, out = self._decode_10(group, out)
            case _:
                out.group = gr
        return out
def test_decoder():
    dec = GroupDecoder()
    from .generator import GroupGenerator
    basic = GroupGenerator.basic(0x3000, tp=True, pty=10)
    ps = GroupGenerator.ps(basic, "radio95 ", 3,ta=True)
    psb = GroupGenerator.ps_b(basic, "radio95 ", 3,ta=True)
    rt = GroupGenerator.rt(basic,"hello!\r ",0)
    rtb = GroupGenerator.rt_b(basic,"hello!\r ",0)
    eccs = []
    for i in range(0xa0, 0xf3):
        eccs.append(GroupGenerator.ecc(basic,i))
    lics = []
    for i in range(0x0, 0x46):
        lics.append(GroupGenerator.lic(basic,i))
    ptyn = GroupGenerator.ptyn(basic, "Test".ljust(8), 0,ab=True)
    print("ps".ljust(6), dec.decode(ps))
    print("psb".ljust(6), dec.decode(psb))
    print("rt".ljust(6), dec.decode(rt))
    print("rtb".ljust(6), dec.decode(rtb))
    for i,ecc in enumerate(eccs):
        print(f"ecc{i}".ljust(6), dec.decode(ecc))
    for i,lic in enumerate(lics):
        print(f"lic{i}".ljust(6), dec.decode(lic))
    print("ptyn".ljust(6), dec.decode(ptyn))